@@ -45,6 +45,7 @@ class IDMLPackage(zipfile.ZipFile):
         self._graphic = None
         self._spreads = None
         self._spreads_objects = None
+        self._last_spread = None
         self._pages = None
         self._backing_story = None
         self._stories = None
@@ -173,7 +174,7 @@ class IDMLPackage(zipfile.ZipFile):
     @property
     def spreads(self):
         if self._spreads is None:
-            spreads = [elt for elt in self.namelist() if re.match(ur"^Spreads/*", elt)]
+            spreads = [elt for elt in self.namelist() if re.match("^Spreads/*", elt)]
             self._spreads = spreads
         return self._spreads
 
@@ -184,6 +185,13 @@ class IDMLPackage(zipfile.ZipFile):
             self._spreads_objects = spreads_objects
         return self._spreads_objects
 
+    @property
+    def last_spread(self):
+        if self._last_spread is None:
+            src = self.designmap.spread_nodes[-1].get("src")
+            self._last_spread = Spread(self, src, self.working_copy_path)
+        return self._last_spread
+
     @property
     def pages(self):
         if self._pages is None:
@@ -205,7 +213,7 @@ class IDMLPackage(zipfile.ZipFile):
     def stories(self):
         if self._stories is None:
             stories = [elt for elt in self.namelist()
-                       if re.match(ur"^%s/*" % STORIES_DIRNAME, elt)]
+                       if re.match("^%s/*" % STORIES_DIRNAME, elt)]
             self._stories = stories
         return self._stories
 
@@ -244,33 +252,17 @@ class IDMLPackage(zipfile.ZipFile):
     def import_xml(self, xml, at):
         """ Reproduce the action «Import XML» on a XML Element in InDesign® Structure. """
 
-        source_node = etree.fromstring(xml)
+        # Python 3 strictly require a bytestring.
+        try:
+            source_node = etree.fromstring(xml)
+        except ValueError:
+            source_node = etree.fromstring(xml.encode("utf-8"))
 
         def _set_content(xpath, element_id, content, story=None):
             story = story or self.get_story_object_by_xpath(xpath)
             story.set_element_content(element_id, content)
             story.synchronize()
 
-        def _set_attributes(xpath, element_id, items):
-            story = self.get_story_object_by_xpath(xpath)
-            story.set_element_attributes(element_id, items)
-            # Image references must be updated in the page item in Spread or Story.
-            if "href" in items:
-                resource_path = items.get("href")
-                element_content_id = self.get_element_content_id_by_xpath(xpath)
-                spread = self.get_spread_object_by_xpath(xpath)
-                if resource_path == "":
-                    story.remove_xml_element_page_items(element_id)
-                    if spread:
-                        spread.remove_page_item(element_content_id, synchronize=True)
-                else:
-                    story.set_element_resource_path(element_content_id, resource_path)
-                    if spread:
-                        spread.set_element_resource_path(element_content_id,
-                                                         resource_path,
-                                                         synchronize=True)
-            story.synchronize()
-
         def _apply_style(style_range_node, style_to_apply_node, applied_style_node):
             """ A style_range_node as an applied_style_node overriden with style_to_apply_node. """
             for attr in ("PointSize", "FontStyle", "HorizontalScale", "Tracking", "FillColor", "Capitalization", "Position"):
@@ -341,7 +333,8 @@ class IDMLPackage(zipfile.ZipFile):
             if parent_style_node is None:
                 return
             # If the parent specify a font face, a font style or a font size and the style_node don't, it is added.
-            for attr in ("PointSize", "FontStyle", "HorizontalScale", "Tracking", "FillColor", "Capitalization", "Position"):
+            for attr in ("PointSize", "FontStyle", "HorizontalScale", "Tracking",
+                         "FillColor", "FillTint", "Capitalization", "Position"):
                 if parent_style_node.get(attr) is not None and (applied_style_node is None or not applied_style_node.get(attr)):
                     style_range_node.set(attr, parent_style_node.get(attr))
 
@@ -352,6 +345,26 @@ class IDMLPackage(zipfile.ZipFile):
                 if attr_node is None and parent_attr_node is not None:
                     properties_element.append(copy.deepcopy(parent_attr_node))
 
+        def _move_siblings_content(at, element_id):
+            """ When new XML elements are inserted, the siblings of the initial <content> (<BR> etc) are
+                repositionned after the last <content> created.
+            """
+            story = self.get_story_object_by_xpath(at)
+            element = story.get_element_by_id(element_id)
+            content_nodes = element.get_element_content_nodes()
+            if len(content_nodes) < 2:
+                return
+
+            first_content_node = content_nodes[0]
+            last_content_node = content_nodes[-1]
+            siblings = [s for s in first_content_node.itersiblings()]
+            if not len(siblings):
+                return
+
+            for sibling in siblings:
+                last_content_node.addnext(sibling)
+            story.synchronize()
+
         def _import_new_node(source_node, at=None, element_id=None, story=None):
             xml_structure_parent_node = self.xml_structure.find("*//*[@Self='%s']" % element_id)
             xml_structure_new_node = etree.Element(source_node.tag)
@@ -389,8 +402,15 @@ class IDMLPackage(zipfile.ZipFile):
             forcecontent = (items.get(FORCECONTENT_TAG) == "true")
             if not ignorecontent_parent_flag or forcecontent:
                 if items:
-                    _set_attributes(at, element_id, items)
-                if not (items.get(SETCONTENT_TAG) == "false"):
+                    self.set_attributes(at, items, element_id)
+                if items.get(SETCONTENT_TAG) == "delete":
+                    local_story = story or self.get_story_object_by_xpath(at)
+                    local_story.remove_element(element_id, synchronize=True)
+                    spread = self.get_spread_object_by_xpath(at)
+                    if spread:
+                        content_id = self.xml_structure.xpath(at)[0].get("XMLContent")
+                        spread.remove_page_item(content_id, synchronize=True)
+                elif not (items.get(SETCONTENT_TAG) == "false"):
                     _set_content(at, element_id, source_node.text or "", story)
 
             ignorecontent = (items.get(IGNORECONTENT_TAG) == "true") or (ignorecontent_parent_flag and not forcecontent)
@@ -404,8 +424,10 @@ class IDMLPackage(zipfile.ZipFile):
                 # we can call a map() on _import_node().
                 # FIXME: what if source_node.text exists ?
                 if destination_node_children_tags == source_node_children_tags:
-                    map(lambda s, d: _import_node(s, at=d, ignorecontent_parent_flag=ignorecontent),
-                        source_node_children, [self.xml_structure_tree.getpath(c) for c in destination_node.iterchildren()])
+                    for s, d in zip(source_node_children, [self.xml_structure_tree.getpath(c) for c in
+                                                           destination_node.iterchildren()]):
+                        _import_node(s, at=d, ignorecontent_parent_flag=ignorecontent)
+
                 # Step-by-step iteration.
                 else:
                     destination_node_child = next(destination_node_children, None)
@@ -419,9 +441,64 @@ class IDMLPackage(zipfile.ZipFile):
                         elif not ignorecontent and source_child.tag in self.style_mapping.character_style_mapping.keys():
                             _import_new_node(source_child, at, element_id)
 
+                    _move_siblings_content(at, element_id)
+
         _import_node(source_node, at)
         return self
 
+    @use_working_copy
+    def import_pdf(self, pdf_path, at, crop="CropContentVisibleLayers"):
+        self.set_attributes(at, {'href': "%s" % pdf_path})
+        spread = self.get_spread_object_by_xpath(at)
+
+        # spread_elt is a <Rectangle> that will host the <PDF>.
+        # The <PDF> element get the id of the <Rectangle>.
+        spread_elt = self.get_spread_elem_by_xpath(at)
+        element_id = self.get_element_content_id_by_xpath(at)
+        spread_elt.set("Self", "%s-old" % element_id)
+        x, y = self.get_elem_point_position(spread_elt)
+        pdf_node = etree.fromstring(f"""<PDF Self="{element_id}" GrayVectorPolicy="IgnoreAll" RGBVectorPolicy="IgnoreAll" CMYKVectorPolicy="IgnoreAll" OverriddenPageItemProps="" LocalDisplaySetting="Default" ImageTypeName="$ID/Adobe Portable Document Format (PDF)" AppliedObjectStyle="ObjectStyle/$ID/[None]" ItemTransform="1 0 0 1 {x} {y}" ParentInterfaceChangeCount="" TargetInterfaceChangeCount="" LastUpdatedInterfaceChangeCount="" HorizontalLayoutConstraints="FlexibleDimension FixedDimension FlexibleDimension" VerticalLayoutConstraints="FlexibleDimension FixedDimension FlexibleDimension" Visible="true" Name="$ID/">
+        <TextWrapPreference Inverse="false" ApplyToMasterPageOnly="false" TextWrapSide="BothSides" TextWrapMode="None">
+            <Properties>
+                <TextWrapOffset Top="0" Left="0" Bottom="0" Right="0" />
+            </Properties>
+            <ContourOption ContourType="SameAsClipping" IncludeInsideEdges="false" ContourPathName="$ID/" />
+        </TextWrapPreference>
+        <PDFAttribute PageNumber="1" PDFCrop="{crop}" TransparentBackground="true" />
+        <MetadataPacketPreference>
+        </MetadataPacketPreference>
+        <Link Self="{element_id}-link" AssetURL="$ID/" AssetID="$ID/" LinkResourceURI="{pdf_path}" LinkResourceFormat="$ID/Adobe Portable Document Format (PDF)" StoredState="Normal" LinkResourceModified="false" LinkObjectModified="false" ShowInUI="true" CanEmbed="true" CanUnembed="true" CanPackage="true" ImportPolicy="NoAutoImport" ExportPolicy="NoAutoExport" />
+        <ClippingPathSettings ClippingType="None" InvertPath="false" IncludeInsideEdges="false" RestrictToFrame="false" UseHighResolutionImage="true" Threshold="25" Tolerance="2" InsetFrame="0" AppliedPathName="$ID/" Index="-1" />
+        <GraphicLayerOption UpdateLinkOption="KeepOverrides" />
+    </PDF>""")
+
+        spread_elt.append(pdf_node)
+        spread.synchronize()
+        return self
+
+    @use_working_copy
+    def set_attributes(self, xpath, items, element_id=None):
+        element_id = element_id or self.xml_structure.xpath(xpath)[0].get("Self")
+        story = self.get_story_object_by_xpath(xpath)
+        story.set_element_attributes(element_id, items)
+        # Image references must be updated in the page item in Spread or Story.
+        if "href" in items:
+            resource_path = items.get("href")
+            element_content_id = self.get_element_content_id_by_xpath(xpath)
+            spread = self.get_spread_object_by_xpath(xpath)
+            if resource_path == "":
+                story.remove_xml_element_page_items(element_id)
+                if spread:
+                    spread.remove_page_item(element_content_id, synchronize=True)
+            else:
+                story.set_element_resource_path(element_content_id, resource_path)
+                if spread:
+                    spread.set_element_resource_path(element_content_id,
+                                                     resource_path,
+                                                     synchronize=True)
+        story.synchronize()
+        return self
+
     def export_as_tree(self):
         """
         tree = {
@@ -471,7 +548,7 @@ class IDMLPackage(zipfile.ZipFile):
             else:
                 # Node without content > `content' is fed with recursive call on childrens.
                 if len(xml_structure_node_children):
-                    content.extend(map(_export_content_as_tree, xml_structure_node_children))
+                    content.extend(list(map(_export_content_as_tree, xml_structure_node_children)))
 
             return tree
 
@@ -482,7 +559,7 @@ class IDMLPackage(zipfile.ZipFile):
         """ Reproduce the action «Export XML» on a XML Element in InDesign® Structure. """
         tree = self.export_as_tree()
         dom = tree_to_etree_dom(tree)
-        return etree.tostring(dom, encoding=encoding, pretty_print=True)
+        return etree.tostring(dom, encoding=encoding, pretty_print=True).decode("utf-8")
 
     @use_working_copy
     def prefix(self, prefix):
@@ -490,7 +567,7 @@ class IDMLPackage(zipfile.ZipFile):
 
         files in ZipFile cannot be renamed or moved so we make a copies of them.
         """
-        if not re.match("^\w+$", prefix):
+        if not re.match(r"^\w+$", prefix):
             raise BaseException("Prefix must be alphanumeric.")
 
         # Change the references inside the file.
@@ -551,7 +628,8 @@ class IDMLPackage(zipfile.ZipFile):
         """Recursively reach the leafs to remove the content. """
         def _remove_content(node):
             if len(node.getchildren()):
-                map(_remove_content, node.iterchildren())
+                for child in node.iterchildren():
+                    _remove_content(child)
             xpath = self.xml_structure_tree.getpath(node)
             element_content_id = self.get_element_content_id_by_xpath(xpath)
 
@@ -567,8 +645,10 @@ class IDMLPackage(zipfile.ZipFile):
         try:
             node = self.xml_structure.xpath(under)[0]
         except IndexError:
-            raise IndexError(u"Cannot remove content under path '%s'. Are you sure the path exists ?" % under)
-        map(_remove_content, node.iterchildren())
+            raise IndexError("Cannot remove content under path '%s'. Are you sure the path exists ?" % under)
+
+        for child in node.iterchildren():
+            _remove_content(child)
         # `under' node may have a reference to its first children in his story.
         story = self.get_story_object_by_xpath(under)
         story.remove_children(node.get("Self"), synchronize=True)
@@ -659,8 +739,8 @@ class IDMLPackage(zipfile.ZipFile):
         at_transform_x, at_transform_y = self.get_elem_translation(at_elem)
 
         only_elem = idml_package.get_spread_elem_by_xpath(only)
-        # <Image> and <EPS> don't store positional informations, so we use the parent <Rectangle>.
-        if only_elem.tag in ["Image", "EPS"]:
+        # <Image>, <EPS> and <PDF> don't store positional informations, so we use the parent <Rectangle>.
+        if only_elem.tag in ["Image", "EPS", "PDF"]:
             only_elem = only_elem.getparent()
         only_rel_pos_x, only_rel_pos_y = idml_package.get_elem_point_position(only_elem, 0)
         only_transform_x, only_transform_y = idml_package.get_elem_translation(only_elem)
@@ -709,7 +789,8 @@ class IDMLPackage(zipfile.ZipFile):
             self.apply_translation_to_element(spread_elt_copy, translation)
             spread_dest_elt.append(spread_elt_copy)
 
-        map(lambda s: _add_spread_element(spread_dest_elt, s), spread_elts_to_add)
+        for elt in spread_elts_to_add:
+            _add_spread_element(spread_dest_elt, elt)
 
         spread_dest.synchronize()
         self.init_lazy_references()
@@ -785,7 +866,7 @@ class IDMLPackage(zipfile.ZipFile):
         if not os.path.exists(stories_dirname):
             os.mkdir(stories_dirname)
         for filename in idml_package.stories_for_node(only):
-            story_cp = open(os.path.join(self.working_copy_path, filename), mode="w+")
+            story_cp = open(os.path.join(self.working_copy_path, filename), mode="wb+")
             story_cp.write(idml_package.open(filename, mode="r").read())
             story_cp.close()
 
@@ -807,7 +888,7 @@ class IDMLPackage(zipfile.ZipFile):
 
     @use_working_copy
     def add_page_from_idml(self, idml_package, page_number, at, only):
-        last_spread = self.spreads_objects[-1]
+        last_spread = self.last_spread
         if last_spread.pages[-1].is_recto:
             last_spread = self.add_new_spread(self.working_copy_path)
 
@@ -831,6 +912,14 @@ class IDMLPackage(zipfile.ZipFile):
         self.init_lazy_references()
         return self
 
+    @use_working_copy
+    def add_note(self, note, author, at, when=None):
+        element_id = self.xml_structure.xpath(at)[0].get("Self")
+        story = self.get_story_object_by_xpath(at)
+        story.add_note(element_id, note, author, when)
+        story.synchronize()
+        return self
+
     @use_working_copy
     def merge_layers(self, with_name=None):
         # Remove all Layer in Designmap but the first.
@@ -861,15 +950,16 @@ class IDMLPackage(zipfile.ZipFile):
     def add_new_spread(self, working_copy_path):
         """Create a new empty Spread in the working copy from the last one. """
 
-        last_spread = self.spreads_objects[-1]
         # TODO : make sure the filename does not exists.
-        new_spread_name = increment_filename(last_spread.name)
+        new_spread_name = increment_filename(self.last_spread.name)
         new_spread_wc_path = os.path.join(working_copy_path, new_spread_name)
         shutil.copy2(
-            os.path.join(working_copy_path, last_spread.name),
+            os.path.join(working_copy_path, self.last_spread.name),
             new_spread_wc_path
         )
+        self._spreads = None
         self._spreads_objects = None
+        self._last_spread = None
 
         new_spread = Spread(self, new_spread_name, working_copy_path)
         new_spread.clear()
@@ -895,7 +985,7 @@ class IDMLPackage(zipfile.ZipFile):
 
     def get_spread_object_by_name(self, name):
         """ - name : 'Spreads/Spread_mainub6.xml' """
-        return filter(lambda s: s.name == name, self.spreads_objects)[0]
+        return next(filter(lambda s: s.name == name, self.spreads_objects))
 
     def get_spread_object_by_xpath(self, xpath):
         elt_id = self.xml_structure.xpath(xpath)[0].get("XMLContent")
@@ -904,7 +994,7 @@ class IDMLPackage(zipfile.ZipFile):
     def get_spread_object_by_id(self, elt_id):
         """elt_id is the `XMLContent' attribute value in the xml_structure (Stories).
 
-        Spread element matches Story one with the ParentStory or the Self attribute value."""
+        Spread element matches Story's one with the ParentStory or the Self attribute value."""
 
         result = None
         for spread in self.spreads_objects:
